import sys
from typing import Union

from .endpoint import Endpoint
from .observer import OptimizationObserver

def ipython_info():
    ip = None
    if 'ipykernel' in sys.modules:
        ip = 'notebook'
    elif 'IPython' in sys.modules:
        ip = 'terminal'
    return ip


def openBrowser(url: str):
    if ipython_info():
        from IPython.display import Javascript, display
        display(Javascript('window.open("{url}");'.format(url=url)))


class BrowserPresenter(OptimizationObserver):
    def __init__(
        self,
        webfrontend_url: str,
    ) -> None:
        self._webfrontend_url = webfrontend_url

    def on_start(self, run_id, *args, **kwargs):
        self.openBrowser(f"run/{run_id}")

    def openBrowser(self, path):
        from urllib.parse import urljoin
        openBrowser(urljoin(self._webfrontend_url, path))

    @classmethod
    def from_endpoint(cls, endpoint: Endpoint):
        host = endpoint.host
        if host.startswith("grpc"):
            www_host = host.replace("grpc", "www")
            return BrowserPresenter(webfrontend_url=f"https://{www_host}:{endpoint.port}")
        elif host == "localhost" and endpoint.port == 9111:
            # development environment
            return BrowserPresenter(webfrontend_url="http://localhost:3000")
        else:
            return None



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union

import grpc

from .grpc.qcalibrate_pb2_grpc import QCalibrateStub as ClientStub
from .auth import GrpcAuth


def _bytes_or_read_file(bytes_or_file: Union[bytes, str]) -> bytes:
    if type(bytes_or_file) is str:
        with open(bytes_or_file, 'rb') as fh:
            return fh.read()
    elif type(bytes_or_file) is bytes:
        return bytes_or_file
    else:
        raise TypeError(f'Expected bytes or string, got {type(bytes_or_file)}')


class Endpoint(ABC):
    """Defines a service endpoint to connect to"""

    def __init__(
            self,
            host: str,
            port: int
    ) -> None:
        """[summary]

        Parameters
        ----------
        host : str
            endpoint host
        port : int
            enpoint port
        """

        super().__init__()
        self._host = host
        self._port = port

    def _get_target(self):
        return f"{self._host}:{self._port}"

    @abstractmethod
    def create_channel(self) -> grpc.Channel:
        """creates a client proxy to communicate with beckend, for internal use only """
        pass

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port


class EndpointSecureKey(Endpoint):
    def __init__(
        self,
        host: str,
        port: int,
        token: str,
        ca_cert: Union[bytes, str] = None,
        compression: grpc.Compression = grpc.Compression.Deflate
    ) -> None:
        """Defines endpoint secured with TLS and a client key

        Parameters
        ----------
        host : str
            endpoint host
        port : int
            enpoint port
        token : str
            client authentication token
        ca_cert : Union[bytes, str], optional
            server CA certificate file content or certificate file name (.pem)
            used to verify server identity in case it uses a self-signed server certificate
            default None
        compression : grpc.Compression, optional
            selects compression algorithm, by default grpc.Compression.Deflate
        """

        super().__init__(host = host, port = port)

        self._ca_cert = _bytes_or_read_file(ca_cert) if ca_cert else None
        self._call_credentials = grpc.metadata_call_credentials(
            GrpcAuth(token=token)
        )
        self._compression = compression

    def create_channel(self) -> grpc.Channel:
        """Internal function to create a gRPC channle

        Returns
        -------
        grpc.Channel
            the created gRPC channel
        """

        credentials = grpc.composite_channel_credentials(
            grpc.ssl_channel_credentials(self._ca_cert),
            self._call_credentials
        )

        channel = grpc.secure_channel(
            self._get_target(), credentials,
            compression=self._compression)

        return channel



import logging
from queue import Queue
from typing import Callable, Generator, Generic, TypeVar

from icecream import ic

TItem = TypeVar('TItem')


class QueueBuffer(Generic[TItem]):
    def __init__(
        self,
        is_end: Callable[[TItem], bool],
        timeout: float = None
    ) -> None:
        """[summary]

        Parameters
        ----------
        is_end : Callable[[TItem], bool]
            check if an item is terminal message
        timeout : float, optional
            Timeout in seconds to wait for a remote reply, by default None - wait forever
        """
        self._queue = Queue()
        self._timeout = timeout
        self._finished = False
        self._is_end = is_end

    def put(self, request: TItem) -> None:
        ic("QueueBuffer.put", request)
        if not self._finished:
            if not request or self._is_end(request):
                self._finished = True
            self._queue.put(request)
        else:
            logging.warn("Send after end %s", request)

    def get(self) -> Generator[TItem, None, None]:
        run = True
        while run:
            next: TItem = self._queue.get(timeout=self._timeout)
            if next:
                ic("QueueBuffer.__next__", next)
                yield next
                if self._is_end(next):
                    run = False
            else:
                run = False



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  Qruise Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Protocol

from numpy import ndarray


@dataclass
class FigureOfMerit:
    """ The result of infidelity (figure of merit evaluation).  
    """
    figure_of_merit: float
    """Pair of a figure of merit value (infidelity measure)
    """
    data: str
    """An arbitrary metadata to be stored in the database, i.e. raw measurement data
    """


@dataclass
class Pulse:
    """Defines ev
    """
    times: ndarray
    """puls value sampling offsets"""
    values: ndarray
    """puls values"""


class EvaluateFigureOfMerit(metaclass=ABCMeta):
    """Abstract class for figure of merit evaluation"""
    """. """

    @abstractmethod
    def evaluate(self, parameters: Dict[str, float], pulses: Dict[str, Pulse]) -> FigureOfMerit:
        """Abstract method for figure of merit evaluation

        Parameters
        ----------
        parameters : Dict[str,float]
            free optimistation parameters
        pulses : Dict[str, Pulse]
            dCRAB pulses 

        Returns
        -------
        FigureOfMerit
            data class of figure of merit and an arbitrary data serialized as a string
        """
        pass

    def close(self):
        """ Invoked on end of an optimization session, can be used to clean up 
        """
        ...


class EvaluateFigureOfMeritFactory(Protocol):
    """abstract figure of merit factory (abstract constructor)
    """
    def __call__(self, *args: Any, **kwds: Any) -> EvaluateFigureOfMerit: ...



from base64 import b64encode, b64decode
from icecream import ic

GRPC_AUTH_HEADER = 'rpc-auth-header'

_ASCII = "ascii"

def splitKeyAndSecret(keyAndSecret):
    (client_key, client_secret) = b64decode(str.encode(keyAndSecret, _ASCII)).decode(_ASCII).split(':', 2)
    return client_key,client_secret

def combineKeyAndSecret(client_key, client_secret):
    return b64encode(f"{client_key}:{client_secret}".encode(_ASCII)).decode(_ASCII)



from typing import Union

from .buffer import QueueBuffer

def bytes_or_read_file(bytes_or_file: Union[bytes, str]) -> bytes:
    if type(bytes_or_file) is str:
        with open(bytes_or_file, 'rb') as fh:
            return fh.read()
    elif type(bytes_or_file) is bytes:
        return bytes_or_file
    else:
        raise TypeError(f'Expected bytes or string, got {type(bytes_or_file)}')

__all__ = [bytes_or_read_file, QueueBuffer]


import logging
from typing import Generic, Iterable, Type, TypeVar, Union

import numpy as np
from cuid import cuid
from icecream import ic
import inspect

from .grpc import qcalibrate_pb2 as pb2
from .grpc.qcalibrate_pb2_grpc import QCalibrateStub
from .utils import QueueBuffer
from .browser import BrowserPresenter
from .endpoint import Endpoint, EndpointSecureKey
from .fom import EvaluateFigureOfMerit, EvaluateFigureOfMeritFactory, Pulse
from .observer import OptimizationObserver, OptimizationResultCollector


def _is_request_end(request: pb2.Request) -> bool:
    return request.HasField('end')


TObserver = TypeVar('TObserver', bound=OptimizationObserver)

OpenBrowserOption = Union[None, bool]

class OptimizationRunner:

    def __init__(
        self,
        stub: QCalibrateStub,
        experiment_id: str,
        evaluate_fom_factory: EvaluateFigureOfMeritFactory,
        process_timeout: float,
        observers: Iterable[OptimizationObserver]
    ) -> None:

        self._stub = stub
        self._experiment_id = experiment_id
        self._evaluate_fom_factory = evaluate_fom_factory
        self._fom: EvaluateFigureOfMerit = None
        self._process_timeout = process_timeout
        self._buffer = QueueBuffer[pb2.Request](
            is_end=_is_request_end,
            timeout=process_timeout)
        self._configuration: pb2.Initialized = None
        self._session_id = cuid()
        self._observers = observers

    def _send(self, request: pb2.Request) -> None:
        ic()
        ic(request)
        self._buffer.put(request)

    def _create_fom(self, configuration: pb2.Initialized) -> None:
        self._configuration = configuration

        for observer in self._observers:
            observer.on_start(
                run_id=configuration.run_id,
                configuration=configuration,
                parameter_metas=configuration.parameter_metas,
                pulse_metas=configuration.pulse_metas)

        arguments = configuration.arguments
        kvargs = {arg.name: getattr(arg, arg.WhichOneof(
            'value')) for arg in arguments}
        ic(kvargs)
        self._fom = self._evaluate_fom_factory(**kvargs)
        self._send(pb2.Request(start=pb2.Start()))

    def _evaluate_fom(self, parameters: pb2.Parameters) -> None:
        try:
            iteration = parameters.iteration
            merged_parameters = {k.name: v for (k, v) in zip(
                self._configuration.parameter_metas, parameters.values)}
            merged_pulses = {k.name: Pulse(np.array(v.times), np.array(v.values)) for (k, v) in zip(
                self._configuration.pulse_metas, parameters.pulses)}
            fom = self._fom.evaluate(
                parameters=merged_parameters, pulses=merged_pulses)

            for observer in self._observers:
                observer.on_iteration(
                    iteration=iteration,
                    parameters=merged_parameters,
                    pulses=merged_pulses,
                    figure_of_merit=fom.figure_of_merit,
                    data=fom.data)

            self._send(pb2.Request(fom=pb2.Fom(
                iteration=iteration,
                figure_of_merit=fom.figure_of_merit, data=fom.data)))

        except Exception as ex:
            logging.exception(ex)
            self._send(pb2.Request(end=pb2.End(
                typ=ex.__class__.__name__, error=str(ex))))

    def _end(self, typ: str = None, error: str = None):
        self._send(pb2.Request(end=pb2.End(typ=typ, error=error)))

    def run(self):
        ic(self._experiment_id, self._session_id)
        self._buffer.put(pb2.Request(initialize=pb2.Initialize(
            experiment_id=self._experiment_id, session_id=self._session_id)))

        response_stream = self._stub.Run(self._buffer.get())
        self._run_loop(response_stream)

    def _run_loop(self, response_stream: Iterable[pb2.Response]) -> None:
        try:
            ic()
            ic(response_stream)
            response: pb2.Response

            for response in response_stream:
                ic()
                ic(response)
                message = response.WhichOneof('message')
                if message == 'initialized':
                    self._create_fom(response.initialized)
                elif message == 'parameters':
                    self._evaluate_fom(response.parameters)
                elif message == 'end':
                    # signal to producer to close
                    self._buffer.put(None)
                    break
                else:
                    raise Exception(message="Invalid message", arg=message)

            for observer in self._observers:
                observer.on_end()

        except Exception as ex:
            logging.exception(ex)
            # ic()
            # ic(traceback.format_exc(ex))
            self._end(typ=ex.__class__.__name__, error=str(ex))  # end cycle
            for observer in self._observers:
                observer.on_error(exception=ex)
        finally:
            if self._fom:
                self._fom.close()


class QOptimizerClient():
    """Interface to an remote QCalibrate optimizer
    """

    def __init__(self, endpoint: Endpoint, remote_timeout: float = None):
        """Constructs optimization clients

        Parameters
        ----------
        endpoint : Endpoint
            client connection endpoint, creates a communication channel
        process_timeout : float, optional
            Timeout in seconds to wait for a remote reply, by default None - wait forever
        """
        self._browser_presenter = BrowserPresenter.from_endpoint(endpoint)
        self._stub = QCalibrateStub(endpoint.create_channel())
        self._process_timeout = remote_timeout

    def run(
            self,
            experiment_id: str,
            evaluate_fom_class: Union[Type[EvaluateFigureOfMerit], EvaluateFigureOfMeritFactory],
            observer: Union[Type[TObserver],TObserver] = OptimizationResultCollector,
            debug=False,
            show_progress: OpenBrowserOption = False,
    ):
        """Runs optimization using a remote optimizer, evaluates figure of merit locally use specified class
        class is instatiated with optional argument  specified in the the experiment configuration 


        Parameters
        ----------
        experiment_id : str
            the experiment id, used to configure the optimizer 
        evaluate_fom_class : Union[Type[EvaluateFom], EvaluateFomFactory]
            class to evaluate figure of merit
        observer : OptimizationObserver
            object to observer and accumulate optimization results, default StoringOptimizationObserver() 
        debug : bool, optional
            Enables addition logs, can help debugging the issues, by default False
        show_progress : bool, optional
            Open browser with run details (if running in IPython notebook), by default False

        Returns
        -------
        TObserver
            Observer 
        """


        if inspect.isclass(observer):
            observer_instance:TObserver = observer()
        else:
            observer_instance:TObserver = observer

        observers = [observer_instance]
        if show_progress and self._browser_presenter:
            observers.append(self._browser_presenter)

        was_enabled = ic.enabled
        if not debug:
            ic.disable()
        try:
            OptimizationRunner(
                stub=self._stub,
                experiment_id=experiment_id,
                evaluate_fom_factory=evaluate_fom_class,
                process_timeout=self._process_timeout,
                observers=observers,
            ).run()
            return observer_instance
        finally:
            if was_enabled:
                ic.enable()


def create_optimizer_client(
        host: str,
        port: int,
        token: str,
        ca_cert: Union[bytes, str] = None,
        remote_timeout: float = None
) -> QOptimizerClient:
    """Creates remote optimizer client

    Parameters
    ----------
    host : str
        endpoint host
    port : int
        enpoint port
    token : str
        client authentication token
    ca_cert : Union[bytes, str], optional
        server CA certificate file content or certificate file name (.pem)
        used to verify server identity in case it uses a self-signed server certificate
        default None
    process_timeout : float, optional
        Timeout in seconds to wait for a remote reply, by default None - wait forever

    Returns
    -------
    QCalibrateClient

    """
    endpoint = EndpointSecureKey(host=host, port=port, token=token, ca_cert=ca_cert)
    return QOptimizerClient(endpoint=endpoint, remote_timeout=remote_timeout)



from qcalibrateremote.grpc import combineKeyAndSecret, splitKeyAndSecret
from cuid import cuid
from icecream import ic

def test_combine_and_split():
    key = cuid()
    secret = cuid()
    ic(key, secret)
    
    keyAndSecret = combineKeyAndSecret(client_key=key, client_secret=secret)
    ic(keyAndSecret)
    (k,s) = splitKeyAndSecret(keyAndSecret)
    assert key == k
    assert secret == s



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qcalibrate.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='qcalibrate.proto',
  package='qcalibrateremote.grpc',
  syntax='proto3',
  serialized_options=b'\n\020eu.qruise.qserveB\017QCalibrateProtoP\001\242\002\002QC',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10qcalibrate.proto\x12\x15qcalibrateremote.grpc\"^\n\x14\x41\x64\x64\x45xperimentRequest\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x15\n\rconfiguration\x18\x05 \x01(\t\",\n\x13\x41\x64\x64\x45xperimentResult\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\"7\n\nInitialize\x12\x12\n\nsession_id\x18\x01 \x01(\t\x12\x15\n\rexperiment_id\x18\x02 \x01(\t\"\xab\x02\n\x0bInitialized\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x0e\n\x06run_id\x18\x02 \x01(\t\x12\x0b\n\x03typ\x18\x03 \x01(\t\x12\x13\n\x0bmodule_name\x18\x04 \x01(\t\x12\x12\n\nclass_name\x18\x05 \x01(\t\x12\x32\n\targuments\x18\x06 \x03(\x0b\x32\x1f.qcalibrateremote.grpc.Argument\x12\x15\n\rconfiguration\x18\x07 \x01(\t\x12=\n\x0fparameter_metas\x18\n \x03(\x0b\x32$.qcalibrateremote.grpc.ParameterMeta\x12\x35\n\x0bpulse_metas\x18\x0b \x03(\x0b\x32 .qcalibrateremote.grpc.PulseMeta\"?\n\tPulseMeta\x12\r\n\x05index\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x15\n\rconfiguration\x18\x03 \x01(\t\"C\n\rParameterMeta\x12\r\n\x05index\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x15\n\rconfiguration\x18\x03 \x01(\t\"f\n\x08\x41rgument\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x05int64\x18\n \x01(\x03H\x00\x12\x10\n\x06\x64ouble\x18\x0b \x01(\x01H\x00\x12\x10\n\x06string\x18\x0c \x01(\tH\x00\x12\x0e\n\x04\x62ool\x18\r \x01(\x08H\x00\x42\x07\n\x05value\"\x07\n\x05Start\"!\n\x03\x45nd\x12\x0b\n\x03typ\x18\x01 \x01(\t\x12\r\n\x05\x65rror\x18\x02 \x01(\t\".\n\x05Pulse\x12\x11\n\x05times\x18\x01 \x03(\x01\x42\x02\x10\x01\x12\x12\n\x06values\x18\x02 \x03(\x01\x42\x02\x10\x01\"a\n\nParameters\x12\x11\n\titeration\x18\x01 \x01(\x05\x12\x12\n\x06values\x18\x02 \x03(\x01\x42\x02\x10\x01\x12,\n\x06pulses\x18\x03 \x03(\x0b\x32\x1c.qcalibrateremote.grpc.Pulse\"?\n\x03\x46om\x12\x11\n\titeration\x18\x01 \x01(\x05\x12\x17\n\x0f\x66igure_of_merit\x18\x02 \x01(\x01\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\t\"\xd2\x01\n\x07Request\x12\x37\n\ninitialize\x18\x01 \x01(\x0b\x32!.qcalibrateremote.grpc.InitializeH\x00\x12-\n\x05start\x18\x02 \x01(\x0b\x32\x1c.qcalibrateremote.grpc.StartH\x00\x12)\n\x03\x66om\x18\x03 \x01(\x0b\x32\x1a.qcalibrateremote.grpc.FomH\x00\x12)\n\x03\x65nd\x18\n \x01(\x0b\x32\x1a.qcalibrateremote.grpc.EndH\x00\x42\t\n\x07message\"\xb4\x01\n\x08Response\x12\x39\n\x0binitialized\x18\x01 \x01(\x0b\x32\".qcalibrateremote.grpc.InitializedH\x00\x12\x37\n\nparameters\x18\x02 \x01(\x0b\x32!.qcalibrateremote.grpc.ParametersH\x00\x12)\n\x03\x65nd\x18\n \x01(\x0b\x32\x1a.qcalibrateremote.grpc.EndH\x00\x42\t\n\x07message2\xc6\x01\n\nQCalibrate\x12j\n\rAddExperiment\x12+.qcalibrateremote.grpc.AddExperimentRequest\x1a*.qcalibrateremote.grpc.AddExperimentResult\"\x00\x12L\n\x03Run\x12\x1e.qcalibrateremote.grpc.Request\x1a\x1f.qcalibrateremote.grpc.Response\"\x00(\x01\x30\x01\x42*\n\x10\x65u.qruise.qserveB\x0fQCalibrateProtoP\x01\xa2\x02\x02QCb\x06proto3'
)




_ADDEXPERIMENTREQUEST = _descriptor.Descriptor(
  name='AddExperimentRequest',
  full_name='qcalibrateremote.grpc.AddExperimentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='qcalibrateremote.grpc.AddExperimentRequest.name', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='qcalibrateremote.grpc.AddExperimentRequest.type', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='qcalibrateremote.grpc.AddExperimentRequest.description', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configuration', full_name='qcalibrateremote.grpc.AddExperimentRequest.configuration', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=43,
  serialized_end=137,
)


_ADDEXPERIMENTRESULT = _descriptor.Descriptor(
  name='AddExperimentResult',
  full_name='qcalibrateremote.grpc.AddExperimentResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='qcalibrateremote.grpc.AddExperimentResult.experiment_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=183,
)


_INITIALIZE = _descriptor.Descriptor(
  name='Initialize',
  full_name='qcalibrateremote.grpc.Initialize',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_id', full_name='qcalibrateremote.grpc.Initialize.session_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='qcalibrateremote.grpc.Initialize.experiment_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=185,
  serialized_end=240,
)


_INITIALIZED = _descriptor.Descriptor(
  name='Initialized',
  full_name='qcalibrateremote.grpc.Initialized',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='qcalibrateremote.grpc.Initialized.experiment_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_id', full_name='qcalibrateremote.grpc.Initialized.run_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='typ', full_name='qcalibrateremote.grpc.Initialized.typ', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module_name', full_name='qcalibrateremote.grpc.Initialized.module_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='class_name', full_name='qcalibrateremote.grpc.Initialized.class_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='qcalibrateremote.grpc.Initialized.arguments', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configuration', full_name='qcalibrateremote.grpc.Initialized.configuration', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameter_metas', full_name='qcalibrateremote.grpc.Initialized.parameter_metas', index=7,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pulse_metas', full_name='qcalibrateremote.grpc.Initialized.pulse_metas', index=8,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=243,
  serialized_end=542,
)


_PULSEMETA = _descriptor.Descriptor(
  name='PulseMeta',
  full_name='qcalibrateremote.grpc.PulseMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='qcalibrateremote.grpc.PulseMeta.index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='qcalibrateremote.grpc.PulseMeta.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configuration', full_name='qcalibrateremote.grpc.PulseMeta.configuration', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=544,
  serialized_end=607,
)


_PARAMETERMETA = _descriptor.Descriptor(
  name='ParameterMeta',
  full_name='qcalibrateremote.grpc.ParameterMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='qcalibrateremote.grpc.ParameterMeta.index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='qcalibrateremote.grpc.ParameterMeta.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configuration', full_name='qcalibrateremote.grpc.ParameterMeta.configuration', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=609,
  serialized_end=676,
)


_ARGUMENT = _descriptor.Descriptor(
  name='Argument',
  full_name='qcalibrateremote.grpc.Argument',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='qcalibrateremote.grpc.Argument.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int64', full_name='qcalibrateremote.grpc.Argument.int64', index=1,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double', full_name='qcalibrateremote.grpc.Argument.double', index=2,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string', full_name='qcalibrateremote.grpc.Argument.string', index=3,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bool', full_name='qcalibrateremote.grpc.Argument.bool', index=4,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='qcalibrateremote.grpc.Argument.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=678,
  serialized_end=780,
)


_START = _descriptor.Descriptor(
  name='Start',
  full_name='qcalibrateremote.grpc.Start',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=782,
  serialized_end=789,
)


_END = _descriptor.Descriptor(
  name='End',
  full_name='qcalibrateremote.grpc.End',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='typ', full_name='qcalibrateremote.grpc.End.typ', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='qcalibrateremote.grpc.End.error', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=791,
  serialized_end=824,
)


_PULSE = _descriptor.Descriptor(
  name='Pulse',
  full_name='qcalibrateremote.grpc.Pulse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='times', full_name='qcalibrateremote.grpc.Pulse.times', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='qcalibrateremote.grpc.Pulse.values', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=872,
)


_PARAMETERS = _descriptor.Descriptor(
  name='Parameters',
  full_name='qcalibrateremote.grpc.Parameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='iteration', full_name='qcalibrateremote.grpc.Parameters.iteration', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='qcalibrateremote.grpc.Parameters.values', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pulses', full_name='qcalibrateremote.grpc.Parameters.pulses', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=874,
  serialized_end=971,
)


_FOM = _descriptor.Descriptor(
  name='Fom',
  full_name='qcalibrateremote.grpc.Fom',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='iteration', full_name='qcalibrateremote.grpc.Fom.iteration', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='figure_of_merit', full_name='qcalibrateremote.grpc.Fom.figure_of_merit', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='qcalibrateremote.grpc.Fom.data', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=973,
  serialized_end=1036,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='qcalibrateremote.grpc.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='initialize', full_name='qcalibrateremote.grpc.Request.initialize', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start', full_name='qcalibrateremote.grpc.Request.start', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fom', full_name='qcalibrateremote.grpc.Request.fom', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end', full_name='qcalibrateremote.grpc.Request.end', index=3,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='message', full_name='qcalibrateremote.grpc.Request.message',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1039,
  serialized_end=1249,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='qcalibrateremote.grpc.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='initialized', full_name='qcalibrateremote.grpc.Response.initialized', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='qcalibrateremote.grpc.Response.parameters', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end', full_name='qcalibrateremote.grpc.Response.end', index=2,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='message', full_name='qcalibrateremote.grpc.Response.message',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1252,
  serialized_end=1432,
)

_INITIALIZED.fields_by_name['arguments'].message_type = _ARGUMENT
_INITIALIZED.fields_by_name['parameter_metas'].message_type = _PARAMETERMETA
_INITIALIZED.fields_by_name['pulse_metas'].message_type = _PULSEMETA
_ARGUMENT.oneofs_by_name['value'].fields.append(
  _ARGUMENT.fields_by_name['int64'])
_ARGUMENT.fields_by_name['int64'].containing_oneof = _ARGUMENT.oneofs_by_name['value']
_ARGUMENT.oneofs_by_name['value'].fields.append(
  _ARGUMENT.fields_by_name['double'])
_ARGUMENT.fields_by_name['double'].containing_oneof = _ARGUMENT.oneofs_by_name['value']
_ARGUMENT.oneofs_by_name['value'].fields.append(
  _ARGUMENT.fields_by_name['string'])
_ARGUMENT.fields_by_name['string'].containing_oneof = _ARGUMENT.oneofs_by_name['value']
_ARGUMENT.oneofs_by_name['value'].fields.append(
  _ARGUMENT.fields_by_name['bool'])
_ARGUMENT.fields_by_name['bool'].containing_oneof = _ARGUMENT.oneofs_by_name['value']
_PARAMETERS.fields_by_name['pulses'].message_type = _PULSE
_REQUEST.fields_by_name['initialize'].message_type = _INITIALIZE
_REQUEST.fields_by_name['start'].message_type = _START
_REQUEST.fields_by_name['fom'].message_type = _FOM
_REQUEST.fields_by_name['end'].message_type = _END
_REQUEST.oneofs_by_name['message'].fields.append(
  _REQUEST.fields_by_name['initialize'])
_REQUEST.fields_by_name['initialize'].containing_oneof = _REQUEST.oneofs_by_name['message']
_REQUEST.oneofs_by_name['message'].fields.append(
  _REQUEST.fields_by_name['start'])
_REQUEST.fields_by_name['start'].containing_oneof = _REQUEST.oneofs_by_name['message']
_REQUEST.oneofs_by_name['message'].fields.append(
  _REQUEST.fields_by_name['fom'])
_REQUEST.fields_by_name['fom'].containing_oneof = _REQUEST.oneofs_by_name['message']
_REQUEST.oneofs_by_name['message'].fields.append(
  _REQUEST.fields_by_name['end'])
_REQUEST.fields_by_name['end'].containing_oneof = _REQUEST.oneofs_by_name['message']
_RESPONSE.fields_by_name['initialized'].message_type = _INITIALIZED
_RESPONSE.fields_by_name['parameters'].message_type = _PARAMETERS
_RESPONSE.fields_by_name['end'].message_type = _END
_RESPONSE.oneofs_by_name['message'].fields.append(
  _RESPONSE.fields_by_name['initialized'])
_RESPONSE.fields_by_name['initialized'].containing_oneof = _RESPONSE.oneofs_by_name['message']
_RESPONSE.oneofs_by_name['message'].fields.append(
  _RESPONSE.fields_by_name['parameters'])
_RESPONSE.fields_by_name['parameters'].containing_oneof = _RESPONSE.oneofs_by_name['message']
_RESPONSE.oneofs_by_name['message'].fields.append(
  _RESPONSE.fields_by_name['end'])
_RESPONSE.fields_by_name['end'].containing_oneof = _RESPONSE.oneofs_by_name['message']
DESCRIPTOR.message_types_by_name['AddExperimentRequest'] = _ADDEXPERIMENTREQUEST
DESCRIPTOR.message_types_by_name['AddExperimentResult'] = _ADDEXPERIMENTRESULT
DESCRIPTOR.message_types_by_name['Initialize'] = _INITIALIZE
DESCRIPTOR.message_types_by_name['Initialized'] = _INITIALIZED
DESCRIPTOR.message_types_by_name['PulseMeta'] = _PULSEMETA
DESCRIPTOR.message_types_by_name['ParameterMeta'] = _PARAMETERMETA
DESCRIPTOR.message_types_by_name['Argument'] = _ARGUMENT
DESCRIPTOR.message_types_by_name['Start'] = _START
DESCRIPTOR.message_types_by_name['End'] = _END
DESCRIPTOR.message_types_by_name['Pulse'] = _PULSE
DESCRIPTOR.message_types_by_name['Parameters'] = _PARAMETERS
DESCRIPTOR.message_types_by_name['Fom'] = _FOM
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AddExperimentRequest = _reflection.GeneratedProtocolMessageType('AddExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADDEXPERIMENTREQUEST,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.AddExperimentRequest)
  })
_sym_db.RegisterMessage(AddExperimentRequest)

AddExperimentResult = _reflection.GeneratedProtocolMessageType('AddExperimentResult', (_message.Message,), {
  'DESCRIPTOR' : _ADDEXPERIMENTRESULT,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.AddExperimentResult)
  })
_sym_db.RegisterMessage(AddExperimentResult)

Initialize = _reflection.GeneratedProtocolMessageType('Initialize', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZE,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Initialize)
  })
_sym_db.RegisterMessage(Initialize)

Initialized = _reflection.GeneratedProtocolMessageType('Initialized', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZED,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Initialized)
  })
_sym_db.RegisterMessage(Initialized)

PulseMeta = _reflection.GeneratedProtocolMessageType('PulseMeta', (_message.Message,), {
  'DESCRIPTOR' : _PULSEMETA,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.PulseMeta)
  })
_sym_db.RegisterMessage(PulseMeta)

ParameterMeta = _reflection.GeneratedProtocolMessageType('ParameterMeta', (_message.Message,), {
  'DESCRIPTOR' : _PARAMETERMETA,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.ParameterMeta)
  })
_sym_db.RegisterMessage(ParameterMeta)

Argument = _reflection.GeneratedProtocolMessageType('Argument', (_message.Message,), {
  'DESCRIPTOR' : _ARGUMENT,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Argument)
  })
_sym_db.RegisterMessage(Argument)

Start = _reflection.GeneratedProtocolMessageType('Start', (_message.Message,), {
  'DESCRIPTOR' : _START,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Start)
  })
_sym_db.RegisterMessage(Start)

End = _reflection.GeneratedProtocolMessageType('End', (_message.Message,), {
  'DESCRIPTOR' : _END,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.End)
  })
_sym_db.RegisterMessage(End)

Pulse = _reflection.GeneratedProtocolMessageType('Pulse', (_message.Message,), {
  'DESCRIPTOR' : _PULSE,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Pulse)
  })
_sym_db.RegisterMessage(Pulse)

Parameters = _reflection.GeneratedProtocolMessageType('Parameters', (_message.Message,), {
  'DESCRIPTOR' : _PARAMETERS,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Parameters)
  })
_sym_db.RegisterMessage(Parameters)

Fom = _reflection.GeneratedProtocolMessageType('Fom', (_message.Message,), {
  'DESCRIPTOR' : _FOM,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Fom)
  })
_sym_db.RegisterMessage(Fom)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Request)
  })
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
  'DESCRIPTOR' : _RESPONSE,
  '__module__' : 'qcalibrate_pb2'
  # @@protoc_insertion_point(class_scope:qcalibrateremote.grpc.Response)
  })
_sym_db.RegisterMessage(Response)


DESCRIPTOR._options = None
_PULSE.fields_by_name['times']._options = None
_PULSE.fields_by_name['values']._options = None
_PARAMETERS.fields_by_name['values']._options = None

_QCALIBRATE = _descriptor.ServiceDescriptor(
  name='QCalibrate',
  full_name='qcalibrateremote.grpc.QCalibrate',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1435,
  serialized_end=1633,
  methods=[
  _descriptor.MethodDescriptor(
    name='AddExperiment',
    full_name='qcalibrateremote.grpc.QCalibrate.AddExperiment',
    index=0,
    containing_service=None,
    input_type=_ADDEXPERIMENTREQUEST,
    output_type=_ADDEXPERIMENTRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Run',
    full_name='qcalibrateremote.grpc.QCalibrate.Run',
    index=1,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_QCALIBRATE)

DESCRIPTOR.services_by_name['QCalibrate'] = _QCALIBRATE

# @@protoc_insertion_point(module_scope)



from grpc import AuthMetadataPlugin
from .grpc import GRPC_AUTH_HEADER


class GrpcAuth(AuthMetadataPlugin):
    """Class implents a metadata plugin adding access token header as call level security

    Parameters
    ----------
    grpc : [type]
        [description]
    """
    def __init__(self, token):
        self._token = token

    def __call__(self, context, callback):
        callback(((GRPC_AUTH_HEADER, self._token),), None)



from posixpath import join
from grpc_tools import protoc
import os

from tempfile import mkstemp
from shutil import move, copymode
import re

def replace(file_path, p, subst):
    #Create temp file
    fh, abs_path = mkstemp()
    with os.fdopen(fh,'w') as new_file:
        with open(file_path) as old_file:
            for line in old_file:
                new_file.write(re.sub(p, subst, line))
    #Copy the file permissions from the old file to the new file
    copymode(file_path, abs_path)
    #Remove original file
    os.remove(file_path)
    #Move new file
    move(abs_path, file_path)

_PROTO_PREFIX = "qcalibrate"

if __name__ == '__main__':
    dir_name = os.path.dirname(os.path.abspath(__file__))

    protoc.main((
        '',
        f'-I{dir_name}',
        f'--python_out={dir_name}',
        f'--grpc_python_out={dir_name}',
        f'{os.path.join(dir_name, f"{_PROTO_PREFIX}.proto")}',
    ))

    files = list(map(lambda f: os.path.join(dir_name,f), filter(lambda s: s.endswith( 'pb2_grpc.py'), os.listdir(dir_name))))

    pattern = re.compile(f'^import {_PROTO_PREFIX}')

    for file in files:
        replace(file, pattern, f'from . import {_PROTO_PREFIX}')





# import sys

# if sys.version_info[:2] >= (3, 8):
#     # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
#     from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
# else:
#     from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

# try:
#     # Change here if project is renamed and does not equal the package name
#     dist_name = __name__
#     __version__ = version(dist_name)
# except PackageNotFoundError:  # pragma: no cover
#     __version__ = "unknown"
# finally:
#     del version, PackageNotFoundError


from .client import QOptimizerClient, create_optimizer_client
from .endpoint import Endpoint, EndpointSecureKey
from .fom import (EvaluateFigureOfMerit, EvaluateFigureOfMeritFactory,
                  FigureOfMerit, Pulse)
from .observer import (IterationResult, OptimizationObserver,
                       RunMetaParameters, ParameterMeta, PulseMeta, OptimizationResultCollector)

from .browser import BrowserPresenter

__all__ = [
    create_optimizer_client,
    BrowserPresenter,
    Endpoint,
    EndpointSecureKey,
    EvaluateFigureOfMerit,
    EvaluateFigureOfMeritFactory,
    FigureOfMerit,
    IterationResult,
    OptimizationObserver, 
    RunMetaParameters,
    ParameterMeta,
    Pulse,
    PulseMeta,
    QOptimizerClient,
    OptimizationResultCollector,
]



from dataclasses import dataclass
from typing import Dict, List

from .fom import Pulse


@dataclass
class IterationResult:
    iteration: int
    parameters: Dict[str, float]
    pulses: Dict[str, Pulse]
    figure_of_merit: float
    data: str


@dataclass
class PulseMeta:
    """Pulse meta parameter"""
    index: int
    """ordinal of a pulse in pulse meta list"""
    name: str
    """name of the pulse"""
    configuration: str
    """JSON serilized pulse optimization meta-parameters like basis, sampling, initial quess and boundaries"""


@dataclass
class ParameterMeta:
    """Parameter optimization meta"""
    index: int
    """ordinal of a parameter in a parametzer list"""
    name: int
    """name of a parameter"""
    configuration: str
    """parameter optimization configuration, like boandaries and initial guess"""


@dataclass
class RunMetaParameters:
    """run attributes"""
    run_id: str
    """database ID"""
    configuration: str
    """JSON serialized optimization parameters"""
    parameter_metas: List[ParameterMeta]
    """List of optimizable parameters"""
    pulse_metas: List[PulseMeta]
    """List of optimizable pulses"""


class OptimizationObserver:
    """Base optimization observer
        don't do anything
    """

    def on_start(
        self,
        run_id,
        configuration: str,
        parameter_metas,
        pulse_metas, **
        kwargs): ...

    def on_iteration(
        self,
        iteration: int,
        parameters: Dict[str, float],
        pulses: Dict[str, Pulse],
        figure_of_merit: float,
        data: str,
        **kwargs): ...

    def on_end(self): ...

    def on_error(self, exception: Exception): ...


class OptimizationResultCollector(OptimizationObserver):
    """Storing optimization observer
    accumulates intermediate iteration results

    Parameters
    ----------
    OptimizationObserver : [type]
        basic type
    """

    def __init__(self) -> None:
        super().__init__()
        self._parameters: RunMetaParameters = None
        self._iterations: List[IterationResult] = []
        self._exception: Exception = None
        self._top: List[IterationResult] = []


    def on_start(self, run_id, configuration: str, parameter_metas, pulse_metas, **kwargs):
        self._parameters = RunMetaParameters(
            run_id=run_id,
            configuration=configuration,
            parameter_metas=parameter_metas,
            pulse_metas=pulse_metas)

    def on_iteration(self,
                     iteration: int,
                     parameters: Dict[str, float],
                     pulses: Dict[str, Pulse],
                     figure_of_merit: float,
                     data: str,
                     **kwargs):
        iteration_result = IterationResult(
            iteration=iteration,
            parameters=parameters,
            pulses=pulses,
            figure_of_merit=figure_of_merit,
            data=data)

        self._iterations.append(iteration_result)

        if len(self._top) == 0:
            self._top.append(iteration_result)            
        else:
            # TODO: implement parametrisable multiple top
            if self._top[0].figure_of_merit > iteration_result.figure_of_merit:
                self._top[0] = iteration_result

    def on_error(self, exception: Exception):
        self._exception = exception

    @property
    def parameters(self):
        """Run parameters"""
        return self._parameters

    @property
    def iterations(self):
        """iteration"""
        return self._iterations

    @property
    def exception(self):
        """exeption in case of error"""
        return self.exception

    @property
    def top(self):
        """best fits, the best first"""
        return self._top
